from musicxml.util.core import cap_first, convert_to_xml_class_name
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement
import xml.etree.ElementTree as ET


class XSDSequence:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self._elements = None
        self.xsd_tree = xsd_tree
        self.parent_container = None

    @property
    def elements(self):
        if not self._elements:
            self._elements = []
            for child in self.xsd_tree.get_children():
                if child.tag == 'element':
                    element = convert_to_xml_class_name(child.name)
                    min_occurrence = child.get_attributes().get('minOccurs')
                    if min_occurrence is None: min_occurrence = '1'
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if max_occurrence is None: max_occurrence = '1'
                    self._elements.append((element, min_occurrence, max_occurrence))

                elif child.tag == 'group':
                    xsd_group_name = 'XSDGroup' + ''.join([cap_first(partial) for partial in child.get_attributes()['ref'].split('-')])
                    elements = eval(xsd_group_name)().sequence.elements
                    min_occurrence = child.get_attributes().get('minOccurs')
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if min_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[1] = min_occurrence
                        elements[0] = tuple(list_el)
                    if max_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[2] = max_occurrence
                        elements[0] = tuple(list_el)
                    self._elements.extend(elements)
                else:
                    raise NotImplementedError(child.tag)
        return self._elements

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'sequence':
            raise ValueError
        self._xsd_tree = value


class XSDChoice:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self.xsd_tree = xsd_tree
        self.parent_container = None

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'choice':
            raise ValueError
        self._xsd_tree = value


class XSDGroup(XSDTreeElement):

    def __init__(self):
        self._sequence = None
        self._name = None
        self.parent_container = None

    @property
    def name(self):
        return self.XSD_TREE.name

    @property
    def sequence(self):
        if not self._sequence:
            for child in self.XSD_TREE.get_children():
                if child.tag == 'sequence':
                    self._sequence = XSDSequence(child)
        return self._sequence


# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_indicators.py
# -----------------------------------------------------


class XSDGroupEditorial(XSDGroup):
    """The editorial group specifies editorial information for a musical element."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="editorial">
    <xs:annotation>
        <xs:documentation>The editorial group specifies editorial information for a musical element.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:group ref="footnote" minOccurs="0" />
        <xs:group ref="level" minOccurs="0" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupEditorialVoice(XSDGroup):
    """The editorial-voice group supports the common combination of editorial and voice information for a musical element."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="editorial-voice">
    <xs:annotation>
        <xs:documentation>The editorial-voice group supports the common combination of editorial and voice information for a musical element.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:group ref="footnote" minOccurs="0" />
        <xs:group ref="level" minOccurs="0" />
        <xs:group ref="voice" minOccurs="0" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupEditorialVoiceDirection(XSDGroup):
    """The editorial-voice-direction group supports the common combination of editorial and voice information for a direction element. It is separate from the editorial-voice element because extensions and restrictions might be different for directions than for the note and forward elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="editorial-voice-direction">
    <xs:annotation>
        <xs:documentation>The editorial-voice-direction group supports the common combination of editorial and voice information for a direction element. It is separate from the editorial-voice element because extensions and restrictions might be different for directions than for the note and forward elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:group ref="footnote" minOccurs="0" />
        <xs:group ref="level" minOccurs="0" />
        <xs:group ref="voice" minOccurs="0" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupFootnote(XSDGroup):
    """The footnote element specifies editorial information that appears in footnotes in the printed score. It is defined within a group due to its multiple uses within the MusicXML schema."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="footnote">
    <xs:annotation>
        <xs:documentation>The footnote element specifies editorial information that appears in footnotes in the printed score. It is defined within a group due to its multiple uses within the MusicXML schema.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="footnote" type="formatted-text" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupLevel(XSDGroup):
    """The level element specifies editorial information for different MusicXML elements. It is defined within a group due to its multiple uses within the MusicXML schema."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="level">
    <xs:annotation>
        <xs:documentation>The level element specifies editorial information for different MusicXML elements. It is defined within a group due to its multiple uses within the MusicXML schema.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="level" type="level" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupStaff(XSDGroup):
    """The staff element is defined within a group due to its use by both notes and direction elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="staff">
    <xs:annotation>
        <xs:documentation>The staff element is defined within a group due to its use by both notes and direction elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="staff" type="xs:positiveInteger">
            <xs:annotation>
                <xs:documentation>Staff assignment is only needed for music notated on multiple staves. Used by both notes and directions. Staff values are numbers, with 1 referring to the top-most staff in a part.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupTuning(XSDGroup):
    """The tuning group contains the sequence of elements common to the staff-tuning and accord elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="tuning">
    <xs:annotation>
        <xs:documentation>The tuning group contains the sequence of elements common to the staff-tuning and accord elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="tuning-step" type="step">
            <xs:annotation>
                <xs:documentation>The tuning-step element is represented like the step element, with a different name to reflect its different function in string tuning.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="tuning-alter" type="semitones" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The tuning-alter element is represented like the alter element, with a different name to reflect its different function in string tuning.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="tuning-octave" type="octave">
            <xs:annotation>
                <xs:documentation>The tuning-octave element is represented like the octave element, with a different name to reflect its different function in string tuning.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupVirtualInstrumentData(XSDGroup):
    """Virtual instrument data can be part of either the score-instrument element at the start of a part, or an instrument-change element within a part."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="virtual-instrument-data">
    <xs:annotation>
        <xs:documentation>Virtual instrument data can be part of either the score-instrument element at the start of a part, or an instrument-change element within a part.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="instrument-sound" type="xs:string" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The instrument-sound element describes the default timbre of the score-instrument. This description is independent of a particular virtual or MIDI instrument specification and allows playback to be shared more easily between applications and libraries.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:choice minOccurs="0">
            <xs:element name="solo" type="empty">
                <xs:annotation>
                    <xs:documentation>The solo element is present if performance is intended by a solo instrument.</xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:element name="ensemble" type="positive-integer-or-empty">
                <xs:annotation>
                    <xs:documentation>The ensemble element is present if performance is intended by an ensemble such as an orchestral section. The text of the ensemble element contains the size of the section, or is empty if the ensemble size is not specified.</xs:documentation>
                </xs:annotation>
            </xs:element>
        </xs:choice>
        <xs:element name="virtual-instrument" type="virtual-instrument" minOccurs="0" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupVoice(XSDGroup):
    """A voice is a sequence of musical events (e.g. notes, chords, rests) that proceeds linearly in time. The voice element is used to distinguish between multiple voices in individual parts. It is defined within a group due to its multiple uses within the MusicXML schema."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="voice">
    <xs:annotation>
        <xs:documentation>A voice is a sequence of musical events (e.g. notes, chords, rests) that proceeds linearly in time. The voice element is used to distinguish between multiple voices in individual parts. It is defined within a group due to its multiple uses within the MusicXML schema.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="voice" type="xs:string" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupClef(XSDGroup):
    """Clefs are represented by a combination of sign, line, and clef-octave-change elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="clef">
    <xs:annotation>
        <xs:documentation>Clefs are represented by a combination of sign, line, and clef-octave-change elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="sign" type="clef-sign">
            <xs:annotation>
                <xs:documentation>The sign element represents the clef symbol.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="line" type="staff-line-position" minOccurs="0">
            <xs:annotation>
                <xs:documentation>Line numbers are counted from the bottom of the staff. They are only needed with the G, F, and C signs in order to position a pitch correctly on the staff. Standard values are 2 for the G sign (treble clef), 4 for the F sign (bass clef), and 3 for the C sign (alto clef). Line values can be used to specify positions outside the staff, such as a C clef positioned in the middle of a grand staff.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="clef-octave-change" type="xs:integer" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The clef-octave-change element is used for transposing clefs. A treble clef for tenors would have a value of -1.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupNonTraditionalKey(XSDGroup):
    """The non-traditional-key group represents a single alteration within a non-traditional key signature. A sequence of these groups makes up a non-traditional key signature"""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="non-traditional-key">
    <xs:annotation>
        <xs:documentation>The non-traditional-key group represents a single alteration within a non-traditional key signature. A sequence of these groups makes up a non-traditional key signature</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="key-step" type="step">
            <xs:annotation>
                <xs:documentation>Non-traditional key signatures are represented using a list of altered tones. The key-step element indicates the pitch step to be altered, represented using the same names as in the step element.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="key-alter" type="semitones">
            <xs:annotation>
                <xs:documentation>Non-traditional key signatures are represented using a list of altered tones. The key-alter element represents the alteration for a given pitch step, represented with semitones in the same manner as the alter element.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="key-accidental" type="key-accidental" minOccurs="0">
            <xs:annotation>
                <xs:documentation>Non-traditional key signatures are represented using a list of altered tones. The key-accidental element indicates the accidental to be displayed in the key signature, represented in the same manner as the accidental element. It is used for disambiguating microtonal accidentals.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupSlash(XSDGroup):
    """The slash group combines elements used for more complete specification of the slash and beat-repeat measure-style elements. They have the same values as the type and dot elements, and define what the beat is for the display of repetition marks. If not present, the beat is based on the current time signature."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="slash">
    <xs:annotation>
        <xs:documentation>The slash group combines elements used for more complete specification of the slash and beat-repeat measure-style elements. They have the same values as the type and dot elements, and define what the beat is for the display of repetition marks. If not present, the beat is based on the current time signature.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:sequence minOccurs="0">
            <xs:element name="slash-type" type="note-type-value">
                <xs:annotation>
                    <xs:documentation>The slash-type element indicates the graphical note type to use for the display of repetition marks.</xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:element name="slash-dot" type="empty" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                    <xs:documentation>The slash-dot element is used to specify any augmentation dots in the note type used to display repetition marks.</xs:documentation>
                </xs:annotation>
            </xs:element>
        </xs:sequence>
        <xs:element name="except-voice" type="xs:string" minOccurs="0" maxOccurs="unbounded">
            <xs:annotation>
                <xs:documentation>The except-voice element is used to specify a combination of slash notation and regular notation. Any note elements that are in voices specified by the except-voice elements are displayed in normal notation, in addition to the slash notation that is always displayed.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupTimeSignature(XSDGroup):
    """Time signatures are represented by the beats element for the numerator and the beat-type element for the denominator."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="time-signature">
    <xs:annotation>
        <xs:documentation>Time signatures are represented by the beats element for the numerator and the beat-type element for the denominator.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="beats" type="xs:string">
            <xs:annotation>
                <xs:documentation>The beats element indicates the number of beats, as found in the numerator of a time signature.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="beat-type" type="xs:string">
            <xs:annotation>
                <xs:documentation>The beat-type element indicates the beat unit, as found in the denominator of a time signature.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupTraditionalKey(XSDGroup):
    """The traditional-key group represents a traditional key signature using the cycle of fifths."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="traditional-key">
    <xs:annotation>
        <xs:documentation>The traditional-key group represents a traditional key signature using the cycle of fifths.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="cancel" type="cancel" minOccurs="0" />
        <xs:element name="fifths" type="fifths" />
        <xs:element name="mode" type="mode" minOccurs="0" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupTranspose(XSDGroup):
    """The transpose group represents what must be added to a written pitch to get a correct sounding pitch."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="transpose">
    <xs:annotation>
        <xs:documentation>The transpose group represents what must be added to a written pitch to get a correct sounding pitch.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="diatonic" type="xs:integer" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The diatonic element specifies the number of pitch steps needed to go from written to sounding pitch. This allows for correct spelling of enharmonic transpositions. This value does not include octave-change values; the values for both elements need to be added to the written pitch to get the correct sounding pitch.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="chromatic" type="semitones">
            <xs:annotation>
                <xs:documentation>The chromatic element represents the number of semitones needed to get from written to sounding pitch. This value does not include octave-change values; the values for both elements need to be added to the written pitch to get the correct sounding pitch.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="octave-change" type="xs:integer" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The octave-change element indicates how many octaves to add to get from written pitch to sounding pitch. The octave-change element should be included when using transposition intervals of an octave or more, and should not be present for intervals of less than an octave.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="double" type="double" minOccurs="0">
            <xs:annotation>
                <xs:documentation>If the double element is present, it indicates that the music is doubled one octave from what is currently written.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupBeatUnit(XSDGroup):
    """The beat-unit group combines elements used repeatedly in the metronome element to specify a note within a metronome mark."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="beat-unit">
    <xs:annotation>
        <xs:documentation>The beat-unit group combines elements used repeatedly in the metronome element to specify a note within a metronome mark.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="beat-unit" type="note-type-value">
            <xs:annotation>
                <xs:documentation>The beat-unit element indicates the graphical note type to use in a metronome mark.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="beat-unit-dot" type="empty" minOccurs="0" maxOccurs="unbounded">
            <xs:annotation>
                <xs:documentation>The beat-unit-dot element is used to specify any augmentation dots for a metronome mark note.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupHarmonyChord(XSDGroup):
    """A harmony element can contain many stacked chords (e.g. V of II). A sequence of harmony-chord groups is used for this type of secondary function, where V of II would be represented by a harmony-chord with a 5 numeral followed by a harmony-chord with a 2 numeral.

A root is a pitch name like C, D, E, while a numeral is a scale degree like 1, 2, 3. The root element is generally used with pop chord symbols, while the numeral element is generally used with classical functional harmony and Nashville numbers. It is an either/or choice to avoid data inconsistency. The function element, which represents Roman numerals with roman numeral text, has been deprecated as of MusicXML 4.0."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="harmony-chord">
    <xs:annotation>
        <xs:documentation>A harmony element can contain many stacked chords (e.g. V of II). A sequence of harmony-chord groups is used for this type of secondary function, where V of II would be represented by a harmony-chord with a 5 numeral followed by a harmony-chord with a 2 numeral.

A root is a pitch name like C, D, E, while a numeral is a scale degree like 1, 2, 3. The root element is generally used with pop chord symbols, while the numeral element is generally used with classical functional harmony and Nashville numbers. It is an either/or choice to avoid data inconsistency. The function element, which represents Roman numerals with roman numeral text, has been deprecated as of MusicXML 4.0.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:choice>
            <xs:element name="root" type="root" />
            <xs:element name="numeral" type="numeral" />
            <xs:element name="function" type="style-text">
                <xs:annotation>
                    <xs:documentation>The function element represents classical functional harmony with an indication like I, II, III rather than C, D, E. It represents the Roman numeral part of a functional harmony rather than the complete function itself. It has been deprecated as of MusicXML 4.0 in favor of the numeral element.</xs:documentation>
                </xs:annotation>
            </xs:element>
        </xs:choice>
        <xs:element name="kind" type="kind" />
        <xs:element name="inversion" type="inversion" minOccurs="0" />
        <xs:element name="bass" type="bass" minOccurs="0" />
        <xs:element name="degree" type="degree" minOccurs="0" maxOccurs="unbounded" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupAllMargins(XSDGroup):
    """The all-margins group specifies both horizontal and vertical margins in tenths."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="all-margins">
    <xs:annotation>
        <xs:documentation>The all-margins group specifies both horizontal and vertical margins in tenths.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:group ref="left-right-margins" />
        <xs:element name="top-margin" type="tenths" />
        <xs:element name="bottom-margin" type="tenths" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupLayout(XSDGroup):
    """The layout group specifies the sequence of page, system, and staff layout elements that is common to both the defaults and print elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="layout">
    <xs:annotation>
        <xs:documentation>The layout group specifies the sequence of page, system, and staff layout elements that is common to both the defaults and print elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="page-layout" type="page-layout" minOccurs="0" />
        <xs:element name="system-layout" type="system-layout" minOccurs="0" />
        <xs:element name="staff-layout" type="staff-layout" minOccurs="0" maxOccurs="unbounded" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupLeftRightMargins(XSDGroup):
    """The left-right-margins group specifies horizontal margins in tenths."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="left-right-margins">
    <xs:annotation>
        <xs:documentation>The left-right-margins group specifies horizontal margins in tenths.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="left-margin" type="tenths" />
        <xs:element name="right-margin" type="tenths" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupDuration(XSDGroup):
    """The duration element is defined within a group due to its uses within the note, figured-bass, backup, and forward elements."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="duration">
    <xs:annotation>
        <xs:documentation>The duration element is defined within a group due to its uses within the note, figured-bass, backup, and forward elements.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="duration" type="positive-divisions">
            <xs:annotation>
                <xs:documentation>Duration is a positive number specified in division units. This is the intended duration vs. notated duration (for instance, differences in dotted notes in Baroque-era music). Differences in duration specific to an interpretation or performance should be represented using the note element's attack and release attributes.

The duration element moves the musical position when used in backup elements, forward elements, and note elements that do not contain a chord child element.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupDisplayStepOctave(XSDGroup):
    """The display-step-octave group contains the sequence of elements used by both the rest and unpitched elements. This group is used to place rests and unpitched elements on the staff without implying that these elements have pitch. Positioning follows the current clef. If percussion clef is used, the display-step and display-octave elements are interpreted as if in treble clef, with a G in octave 4 on line 2."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="display-step-octave">
    <xs:annotation>
        <xs:documentation>The display-step-octave group contains the sequence of elements used by both the rest and unpitched elements. This group is used to place rests and unpitched elements on the staff without implying that these elements have pitch. Positioning follows the current clef. If percussion clef is used, the display-step and display-octave elements are interpreted as if in treble clef, with a G in octave 4 on line 2.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="display-step" type="step" />
        <xs:element name="display-octave" type="octave" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupFullNote(XSDGroup):
    """The full-note group is a sequence of the common note elements between cue/grace notes and regular (full) notes: pitch, chord, and rest information, but not duration (cue and grace notes do not have duration encoded). Unpitched elements are used for unpitched percussion, speaking voice, and other musical elements lacking determinate pitch."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="full-note">
    <xs:annotation>
        <xs:documentation>The full-note group is a sequence of the common note elements between cue/grace notes and regular (full) notes: pitch, chord, and rest information, but not duration (cue and grace notes do not have duration encoded). Unpitched elements are used for unpitched percussion, speaking voice, and other musical elements lacking determinate pitch.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="chord" type="empty" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The chord element indicates that this note is an additional chord tone with the preceding note.

The duration of a chord note does not move the musical position within a measure. That is done by the duration of the first preceding note without a chord element. Thus the duration of a chord note cannot be longer than the preceding note.
							
In most cases the duration will be the same as the preceding note. However it can be shorter in situations such as multiple stops for string instruments.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:choice>
            <xs:element name="pitch" type="pitch" />
            <xs:element name="unpitched" type="unpitched" />
            <xs:element name="rest" type="rest" />
        </xs:choice>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupMusicData(XSDGroup):
    """The music-data group contains the basic musical data that is either associated with a part or a measure, depending on whether the partwise or timewise hierarchy is used."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="music-data">
    <xs:annotation>
        <xs:documentation>The music-data group contains the basic musical data that is either associated with a part or a measure, depending on whether the partwise or timewise hierarchy is used.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:choice minOccurs="0" maxOccurs="unbounded">
            <xs:element name="note" type="note" />
            <xs:element name="backup" type="backup" />
            <xs:element name="forward" type="forward" />
            <xs:element name="direction" type="direction" />
            <xs:element name="attributes" type="attributes" />
            <xs:element name="harmony" type="harmony" />
            <xs:element name="figured-bass" type="figured-bass" />
            <xs:element name="print" type="print" />
            <xs:element name="sound" type="sound" />
            <xs:element name="listening" type="listening" />
            <xs:element name="barline" type="barline" />
            <xs:element name="grouping" type="grouping" />
            <xs:element name="link" type="link" />
            <xs:element name="bookmark" type="bookmark" />
        </xs:choice>
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupPartGroup(XSDGroup):
    """The part-group element is defined within a group due to its multiple uses within the part-list element."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="part-group">
    <xs:annotation>
        <xs:documentation>The part-group element is defined within a group due to its multiple uses within the part-list element.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="part-group" type="part-group" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupScoreHeader(XSDGroup):
    """The score-header group contains basic score metadata about the work and movement, score-wide defaults for layout and fonts, credits that appear on the first or following pages, and the part list."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="score-header">
    <xs:annotation>
        <xs:documentation>The score-header group contains basic score metadata about the work and movement, score-wide defaults for layout and fonts, credits that appear on the first or following pages, and the part list.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="work" type="work" minOccurs="0" />
        <xs:element name="movement-number" type="xs:string" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The movement-number element specifies the number of a movement.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="movement-title" type="xs:string" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The movement-title element specifies the title of a movement, not including its number.</xs:documentation>
            </xs:annotation>
        </xs:element>
        <xs:element name="identification" type="identification" minOccurs="0" />
        <xs:element name="defaults" type="defaults" minOccurs="0" />
        <xs:element name="credit" type="credit" minOccurs="0" maxOccurs="unbounded" />
        <xs:element name="part-list" type="part-list" />
    </xs:sequence>
</xs:group>
"""
                                     ))


class XSDGroupScorePart(XSDGroup):
    """The score-part element is defined within a group due to its multiple uses within the part-list element."""

    XSD_TREE = XSDTree(ET.fromstring("""
<xs:group xmlns:xs="http://www.w3.org/2001/XMLSchema" name="score-part">
    <xs:annotation>
        <xs:documentation>The score-part element is defined within a group due to its multiple uses within the part-list element.</xs:documentation>
    </xs:annotation>
    <xs:sequence>
        <xs:element name="score-part" type="score-part">
            <xs:annotation>
                <xs:documentation>Each MusicXML part corresponds to a track in a Standard MIDI Format 1 file. The score-instrument elements are used when there are multiple instruments per track. The midi-device element is used to make a MIDI device or port assignment for the given track. Initial midi-instrument assignments may be made here as well.</xs:documentation>
            </xs:annotation>
        </xs:element>
    </xs:sequence>
</xs:group>
"""
                                     ))


__all__ = ['XSDSequence', 'XSDChoice', 'XSDGroup', 'XSDGroupEditorial', 'XSDGroupEditorialVoice', 'XSDGroupEditorialVoiceDirection',
           'XSDGroupFootnote', 'XSDGroupLevel', 'XSDGroupStaff', 'XSDGroupTuning', 'XSDGroupVirtualInstrumentData', 'XSDGroupVoice',
           'XSDGroupClef', 'XSDGroupNonTraditionalKey', 'XSDGroupSlash', 'XSDGroupTimeSignature', 'XSDGroupTraditionalKey',
           'XSDGroupTranspose', 'XSDGroupBeatUnit', 'XSDGroupHarmonyChord', 'XSDGroupAllMargins', 'XSDGroupLayout',
           'XSDGroupLeftRightMargins', 'XSDGroupDuration', 'XSDGroupDisplayStepOctave', 'XSDGroupFullNote', 'XSDGroupMusicData',
           'XSDGroupPartGroup', 'XSDGroupScoreHeader', 'XSDGroupScorePart']



import copy
import xml.etree.ElementTree as ET
from contextlib import redirect_stdout
from pathlib import Path
from string import Template

from musicxml.generate_classes.utils import musicxml_xsd_et_root, ns
from musicxml.util.core import convert_to_xml_class_name, convert_to_xsd_class_name
from musicxml.xsd.xsdtree import XSDTree
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdcomplextype import __all__ as all_complex_types
from musicxml.xsd.xsdsimpletype import *

default_path = Path(__file__).parent / 'defaults' / 'xmlelement.py'
target_path = Path(__file__).parent.parent / 'xmlelement' / 'xmlelement.py'

template_string = """
class $class_name($base_classes):
    
    TYPE = $xsd_type
    XSD_TREE = XSDTree(ET.fromstring(\"\"\"
$xsd_string
\"\"\"
                                     ))

    @property
    def __doc__(self):
        if self.TYPE.XSD_TREE.is_complex_type:
            return self.TYPE.__doc__
        else:
            return self.XSD_TREE.get_doc()
"""

typed_elements = list(
    dict.fromkeys(
        [
            (node.attrib['name'], node.attrib['type']) for node in musicxml_xsd_et_root.iter() if node.tag == f'{ns}element' and
                                                                                                  node.attrib.get('type') is not None
        ]
    )
)

xml_element_class_names = ['XMLScorePartwise', 'XMLPart', 'XMLMeasure', 'XMLDirective']


def element_class_as_string(element_):
    found_et_xml = musicxml_xsd_et_root.find(f".//{{*}}element[@name='{element_[0]}'][@type='{element_[1]}']")
    copied_el = copy.deepcopy(found_et_xml)
    if copied_el.attrib.get('minOccurs'):
        copied_el.attrib.pop('minOccurs')
    if copied_el.attrib.get('maxOccurs'):
        copied_el.attrib.pop('maxOccurs')
    xsd_tree = XSDTree(copied_el)
    class_name = convert_to_xml_class_name(xsd_tree.name)
    xml_element_class_names.append(class_name)
    try:
        xsd_type = convert_to_xsd_class_name(xsd_tree.get_attributes()['type'], 'complex_type')
        assert xsd_type in all_complex_types
    except (ValueError, AssertionError):
        xsd_type = convert_to_xsd_class_name(xsd_tree.get_attributes()['type'], 'simple_type')
    base_classes = ('XMLElement',)
    ET.indent(found_et_xml, space='    '),
    xsd_string = ET.tostring(found_et_xml, encoding='unicode').strip()
    t = Template(template_string).substitute(class_name=class_name, base_classes=', '.join(base_classes), xsd_type=xsd_type,
                                             xsd_string=xsd_string)
    return t


with open(target_path, 'w+') as f:
    with open(default_path, 'r') as default:
        with redirect_stdout(f):
            print(default.read())
    with redirect_stdout(f):
        for element in typed_elements:
            print(element_class_as_string(element))
        print(f'__all__={xml_element_class_names}')



from contextlib import redirect_stdout
from pathlib import Path
from string import Template
import xml.etree.ElementTree as ET

from musicxml.generate_classes.utils import get_complex_type_all_base_classes, get_all_et_elements
from musicxml.xsd.xsdtree import XSDTree

sources_path = Path(__file__).parent / 'musicxml_4_0.xsd'
default_path = Path(__file__).parent / 'defaults' / 'xsdcomplextype.py'
target_path = Path(__file__).parent.parent / 'xsd' / 'xsdcomplextype.py'

template_string = """
class $class_name($base_classes):
    \"\"\"$doc\"\"\"
    
    _SIMPLE_CONTENT = $simple_content
    
    XSD_TREE = XSDTree(ET.fromstring(\"\"\"
$xsd_string
\"\"\"
                                     ))
"""

xsd_complex_types = ['XSDComplexType', 'XSDComplexTypeScorePartwise', 'XSDComplexTypePart', 'XSDComplexTypeMeasure',
                     'XSDComplexTypeDirective', 'XSDComplexTypeNote']


def complex_type_class_as_string(complex_type_):
    xsd_tree = XSDTree(complex_type_)
    class_name = xsd_tree.xsd_element_class_name
    xsd_complex_types.append(class_name)
    base_class_names = []
    simple_content = None
    for cls_name in get_complex_type_all_base_classes(xsd_tree):
        if cls_name.startswith('XSDSimpleType'):
            if simple_content is not None:
                raise NotImplementedError('More than one Simple Type as base class.')
            simple_content = cls_name
        else:
            base_class_names.append(cls_name)

    doc = xsd_tree.get_doc()
    ET.indent(complex_type_, space='    '),
    xsd_string = ET.tostring(complex_type_, encoding='unicode').strip()
    if not doc:
        doc = ""
    t = Template(template_string).substitute(class_name=class_name, base_classes=', '.join(base_class_names), simple_content=simple_content,
                                             doc=doc, xsd_string=xsd_string)
    return t


all_complex_type_et_elements = [ct for ct in get_all_et_elements(sources_path, 'complexType') if ct.attrib['name'] != 'note']

with open(target_path, 'w+') as f:
    with open(default_path, 'r') as default:
        with redirect_stdout(f):
            print(default.read())
    with redirect_stdout(f):
        for complex_type in all_complex_type_et_elements:
            print(complex_type_class_as_string(complex_type))
        print(f'__all__={xsd_complex_types}')



import datetime
from pathlib import Path

from musicxml.parser.parser import parse_musicxml

start_reading = datetime.datetime.now()
print("start reading score")
score = parse_musicxml(Path(__file__).parent / 'parser_test.xml')
start_writing = datetime.datetime.now()
print(f"start writing score :{start_writing - start_reading}")
score.write(Path(__file__).parent / 'parser_test_create.xml')
end = datetime.datetime.now()
print(f"end writing score :{end - start_writing}")





from unittest import TestCase

from musicxml.tree.tree import Tree, ChildNotFoundError


class A(Tree):
    def __init__(self, name, parent=None, *args, **keyword):
        super().__init__(*args, **keyword)
        self._children = []
        self._parent = parent
        self.name = name

    def _check_child_to_be_added(self, child):
        if not isinstance(child, self.__class__):
            raise TypeError

    def add_child(self, name):
        child = type(self)(parent=self, name=name)
        return super().add_child(child)


class TestTree(TestCase):
    def setUp(self) -> None:
        self.root = A(name='root')
        self.child1 = self.root.add_child('child1')
        self.child2 = self.root.add_child('child2')
        self.child3 = self.root.add_child('child3')
        self.child4 = self.root.add_child('child4')
        self.grandchild1 = self.child2.add_child('grandchild1')
        self.grandchild2 = self.child2.add_child('grandchild2')
        self.grandchild3 = self.child4.add_child('grandchild3')
        self.greatgrandchild1 = self.grandchild2.add_child('greatgrandchild1')

    def test_get_root(self):
        assert self.greatgrandchild1.get_root() == self.root
        assert self.child4.get_root() == self.root

    def test_is_leaf(self):
        assert self.greatgrandchild1.is_leaf is True
        assert self.child4.is_leaf is False
        assert self.child3.is_leaf is True

    def test_traverse(self):
        assert list(self.root.traverse()) == [self.root, self.child1, self.child2, self.grandchild1, self.grandchild2,
                                              self.greatgrandchild1, self.child3, self.child4, self.grandchild3]

    def test_iterate_leaves(self):
        assert list(self.root.iterate_leaves()) == [self.child1, self.grandchild1, self.greatgrandchild1,
                                                    self.child3, self.grandchild3]

    def test_get_layer_number(self):
        assert [node.get_layer_number() for node in self.root.traverse()] == [0, 1, 1, 2, 2, 3, 1, 1, 2]

    def test_tree_repr(self):
        expected = """root
    child1
    child2
        grandchild1
        grandchild2
            greatgrandchild1
    child3
    child4
        grandchild3
"""
        assert self.root.tree_representation(lambda x: x.name) == expected

    def test_level(self):
        assert self.greatgrandchild1.level == 3
        assert self.grandchild2.level == 2
        assert self.child4.level == 1
        assert self.root.level == 0

    def test_reversed_path_to_root(self):
        assert list(self.greatgrandchild1.reversed_path_to_root()) == [self.greatgrandchild1, self.grandchild2, self.child2, self.root]

    def test_remove_child(self):
        with self.assertRaises(ChildNotFoundError):
            self.child2.remove(self.child1)

        self.child2.remove(self.grandchild2)
        assert self.child2.get_children() == [self.grandchild1]
        assert self.grandchild2.get_parent() is None
        assert self.greatgrandchild1.get_parent() == self.grandchild2

    def test_get_coordinates(self):
        assert self.greatgrandchild1.get_coordinates_in_tree() == '2.2.1'
        assert self.child2.get_coordinates_in_tree() == '2'
        assert self.root.get_coordinates_in_tree() == '0'

    def test_replace_child(self):
        self.child2.replace_child(self.grandchild1, A(name='new_grand_child'))
        assert [ch.name for ch in self.child2.get_children()] == ['new_grand_child', 'grandchild2']
        self.child1.add_child('grandchild')
        self.child1.add_child('grandchild')
        new = A(name='other_new_grand_child')
        self.child1.replace_child(lambda x: x.name == 'grandchild', new, 1)
        assert new.get_parent() == self.child1
        assert [ch.name for ch in self.child1.get_children()] == ['grandchild', 'other_new_grand_child']
        with self.assertRaises(ValueError):
            self.child2.replace_child(None, None)
        with self.assertRaises(TypeError):
            self.root.replace_child(self.child1, 34)



import copy
import xml.etree.ElementTree as ET
from typing import Optional

from musicxml.exceptions import XSDWrongAttribute, XSDAttributeRequiredException, XMLElementChildrenRequired
from musicxml.generate_classes.utils import musicxml_xsd_et_root, ns
from musicxml.tree.tree import Tree
from musicxml.util.core import cap_first, replace_key_underline_with_hyphen
from musicxml.xmlelement.containers import containers
from musicxml.xmlelement.exceptions import XMLElementCannotHaveChildrenError
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdsimpletype import *
from musicxml.xsd.xsdtree import XSDTree


class XMLElement(Tree):
    PROPERTIES = {'compact_repr', 'is_leaf', 'level', 'attributes', 'child_container_tree', 'possible_children_names',
                  'et_xml_element', 'name', 'type_', 'value', 'parent_xsd_element'}
    TYPE = None
    XSD_TREE: Optional[XSDTree] = None

    def __init__(self, value=None, **kwargs):
        self._type = None
        super().__init__()
        self._value = None
        self._attributes = {}
        self._et_xml_element = None
        self._child_container_tree = None
        self._unordered_children = []
        self.value = value
        self._set_attributes(kwargs)

        self._create_child_container_tree()

    def _check_attribute(self, name, value):
        attributes = self.TYPE.get_xsd_attributes()
        allowed_attributes = [attribute.name for attribute in attributes]
        if name not in [attribute.name for attribute in self.TYPE.get_xsd_attributes()]:
            raise XSDWrongAttribute(f"{self.__class__.__name__} has no attribute {name}. Allowed attributes are: {allowed_attributes}")
        for attribute in attributes:
            if attribute.name == name:
                return attribute(value)

    def _check_child_to_be_added(self, child):
        if not isinstance(child, XMLElement):
            raise TypeError

    def _check_required_attributes(self):
        if self.TYPE.XSD_TREE.is_complex_type:
            required_attributes = [attribute for attribute in self.TYPE.get_xsd_attributes() if attribute.is_required]
            for required_attribute in required_attributes:
                if required_attribute.name not in self.attributes:
                    raise XSDAttributeRequiredException(f"{self.__class__.__name__} requires attribute: {required_attribute.name}")

    def _convert_attribute_to_child(self, name, value):
        if not name.startswith('xml_'):
            raise NameError
        child_name = name.replace('xml_', '')

        if '-'.join(child_name.split('_')) not in self.possible_children_names:
            raise NameError

        child_class_name = 'XML' + ''.join([cap_first(partial) for partial in child_name.split('_')])
        child_class = eval(child_class_name)

        found_child = self.find_child(child_class_name)
        if isinstance(value, child_class):
            if found_child:
                self.replace_child(found_child, value)
            else:
                self.add_child(value)
        elif value is None:
            if found_child:
                self.remove(found_child)
        else:
            if found_child:
                found_child.value = value
            else:
                self.add_child(child_class(value))

    def _create_child_container_tree(self):
        try:
            if self.TYPE.XSD_TREE.is_complex_type:
                self._child_container_tree = copy.copy(containers[self.TYPE.__name__])
                self._child_container_tree._parent_xml_element = self
        except KeyError:
            pass

    def _create_et_xml_element(self):
        self._et_xml_element = ET.Element(self.name, {k: str(v) for k, v in self.attributes.items()})
        if self.value is not None:
            self._et_xml_element.text = str(self.value)
        for child in self.get_children():
            self._et_xml_element.append(child.et_xml_element)
        ET.indent(self._et_xml_element, space="    ", level=self.level)

    def _final_checks(self, intelligent_choice=False):
        if self._child_container_tree:
            required_children = self._child_container_tree.get_required_element_names(intelligent_choice=intelligent_choice)
            if required_children:
                raise XMLElementChildrenRequired(f"{self.__class__.__name__} requires at least following children: {required_children}")

        self._check_required_attributes()

        for child in self.get_children():
            child._final_checks(intelligent_choice=intelligent_choice)

    def _get_attributes_error_message(self, wrong_name):
        attributes = self.TYPE.get_xsd_attributes()
        allowed_attributes = [attribute.name for attribute in attributes]
        return f"{self.__class__.__name__} has no attribute {wrong_name}. Allowed attributes are: " \
               f"{sorted(allowed_attributes)} or possible " \
               f"children as attributes: {sorted(['xml_' + '_'.join(ch.split('-')) for ch in self.possible_children_names])}"

    def _set_attributes(self, val):
        if not val:
            return

        if self.TYPE.XSD_TREE.is_simple_type:
            if val:
                raise XSDWrongAttribute(f'{self.__class__.__name__} has no attributes.')

        elif not isinstance(val, dict):
            raise TypeError

        new_attributes = replace_key_underline_with_hyphen(dict_=val)
        none_values_dict = {k: v for k, v in new_attributes.items() if v is None}
        for key in none_values_dict:
            new_attributes.pop(key)
            try:
                self.attributes.pop(key)
            except KeyError:
                pass
        for key in new_attributes:
            self._check_attribute(key, new_attributes[key])
        self._attributes = {**self._attributes, **new_attributes}

    @property
    def attributes(self):
        return self._attributes

    @property
    def child_container_tree(self):
        return self._child_container_tree

    @property
    def et_xml_element(self):
        self._create_et_xml_element()
        return self._et_xml_element

    @property
    def name(self):
        return self.XSD_TREE.get_attributes()['name']

    @property
    def possible_children_names(self):
        if not self.child_container_tree:
            return {}
        else:
            return {leaf.content.name for leaf in self.child_container_tree.iterate_leaves()}

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if val is not None:
            self.TYPE(val)
            self._value = val

    @classmethod
    def get_xsd(cls):
        return cls.XSD_TREE.get_xsd()

    @classmethod
    def get_class_name(cls):
        return cls.__name__

    def add_child(self, child: 'XMLElement', forward=None) -> 'XMLElement':
        if not self._child_container_tree:
            raise XMLElementCannotHaveChildrenError()
        self._child_container_tree.add_element(child, forward)
        self._unordered_children.append(child)
        return child

    def get_children(self, ordered=True):
        if ordered is False:
            return self._unordered_children
        if self._child_container_tree:
            return [xml_element for leaf in self._child_container_tree.iterate_leaves() for xml_element in leaf.content.xml_elements if
                    leaf.content.xml_elements]
        else:
            return []

    def find_child(self, name, ordered=False):
        if isinstance(name, type):
            name = name.__name__
        for ch in self.get_children(ordered=ordered):
            if ch.__class__.__name__ == name:
                return ch

    def find_children(self, name, ordered=False):
        if isinstance(name, type):
            name = name.__name__
        return [ch for ch in self.get_children(ordered=ordered) if ch.__class__.__name__ == name]

    def remove(self, child):
        self._unordered_children.remove(child)

        parent_container = child.parent_xsd_element.parent_container.get_parent()
        if parent_container.chosen_child == child.parent_xsd_element.parent_container:
            parent_container.chosen_child = None
            parent_container.requirements_not_fulfilled = True
        child.parent_xsd_element.xml_elements.remove(child)
        child.parent_xsd_element = None
        del child

    def replace_child(self, old, new, index: int = 0) -> None:
        """
        :param old: child or function
        :param new: child
        :param index: index of old in list of old appearances
        :return: None
        """
        if hasattr(old, '__call__'):
            list_of_olds = [ch for ch in self.get_children(ordered=True) if old(ch)]
        else:
            list_of_olds = [ch for ch in self.get_children(ordered=True) if ch == old]

        if not list_of_olds:
            raise ValueError(f"{old} not in list.")
        self._check_child_to_be_added(new)
        old_index = self._unordered_children.index(list_of_olds[index])
        old_child = self._unordered_children[old_index]
        self._unordered_children.remove(old_child)
        self._unordered_children.insert(old_index, new)

        parent_xsd_element = old_child.parent_xsd_element
        new.parent_xsd_element = parent_xsd_element
        parent_xsd_element._xml_elements = [new if el == old_child else el for el in parent_xsd_element.xml_elements]

    def to_string(self, intelligent_choice=False) -> str:
        self._final_checks(intelligent_choice=intelligent_choice)
        self._create_et_xml_element()

        return ET.tostring(self.et_xml_element, encoding='unicode') + '\n'

    def __setattr__(self, key, value):
        if key[0] == '_' or key in self.PROPERTIES:
            if key == 'value' and value is not None:
                try:
                    self._set_attributes({key: value})
                except XSDWrongAttribute:
                    super().__setattr__(key, value)
            else:
                super().__setattr__(key, value)
        elif key.startswith('xml_'):
            try:
                self._convert_attribute_to_child(name=key, value=value)
            except NameError:
                raise AttributeError(self._get_attributes_error_message(key))
        else:
            try:
                self._set_attributes({key: value})
            except XSDWrongAttribute:
                raise AttributeError(self._get_attributes_error_message(key))

    def __getattr__(self, item):
        try:
            return self.attributes['-'.join(item.split('_'))]
        except KeyError:
            attributes = self.TYPE.get_xsd_attributes()
            allowed_attributes = ['_'.join(attribute.name.split('-')) for attribute in attributes]
            if item in allowed_attributes:
                return None
            else:
                if item.startswith('xml'):
                    child_name = item.replace('xml_', '')
                    for child in self.get_children(ordered=False):
                        if child.name == '-'.join(child_name.split('_')):
                            return child
                    if '-'.join(child_name.split('_')) in self.possible_children_names:
                        return None
                raise AttributeError(self._get_attributes_error_message(item))


# xml score partwise
xsd_tree_score_partwise_part = XSDTree(musicxml_xsd_et_root.find(f".//{ns}element[@name='score-partwise']"))
"""
<xs:element name="score-partwise" block="extension substitution" final="#all">
    <xs:annotation>
        <xs:documentation>The score-partwise element is the root element for a partwise MusicXML score. It includes a score-header group followed by a series of parts with measures inside. The document-attributes attribute group includes the version attribute.</xs:documentation>
    </xs:annotation>
    <xs:complexType>
        <xs:sequence>
            <xs:group ref="score-header"/>
            <xs:element name="part" maxOccurs="unbounded">
                <xs:complexType>
                    <xs:sequence>
                        <xs:element name="measure" maxOccurs="unbounded">
                            <xs:complexType>
                                <xs:group ref="music-data"/>
                                <xs:attributeGroup ref="measure-attributes"/>
                            </xs:complexType>
                        </xs:element>
                    </xs:sequence>
                    <xs:attributeGroup ref="part-attributes"/>
                </xs:complexType>
            </xs:element>
        </xs:sequence>
        <xs:attributeGroup ref="document-attributes"/>
    </xs:complexType>
</xs:element>
"""


class XMLScorePartwise(XMLElement):
    TYPE = XSDComplexTypeScorePartwise
    XSD_TREE = XSDTree(musicxml_xsd_et_root.find(f".//{ns}element[@name='score-partwise']"))

    def write(self, path, intelligent_choice=False):
        with open(path, 'w') as file:
            file.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n')
            file.write(self.to_string(intelligent_choice=intelligent_choice))

    @property
    def __doc__(self):
        return self.XSD_TREE.get_doc()


class XMLPart(XMLElement):
    TYPE = XSDComplexTypePart
    XSD_TREE = XSDTree(musicxml_xsd_et_root.findall(f".//{ns}element[@name='score-partwise']//{ns}element")[0])

    @property
    def __doc__(self):
        return self.XSD_TREE.get_doc()


class XMLMeasure(XMLElement):
    TYPE = XSDComplexTypeMeasure
    XSD_TREE = XSDTree(musicxml_xsd_et_root.findall(f".//{ns}element[@name='score-partwise']//{ns}element")[1])

    @property
    def __doc__(self):
        return self.XSD_TREE.get_doc()


class XMLDirective(XMLElement):
    TYPE = XSDComplexTypeDirective
    XSD_TREE = XSDTree(musicxml_xsd_et_root.find(".//{*}complexType[@name='attributes']//{*}element[@name='directive']"))

    @property
    def __doc__(self):
        return self.XSD_TREE.get_doc()
# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_xml_elements.py
# -----------------------------------------------------



from musicxml.util.core import convert_to_xml_class_name
import xml.etree.ElementTree as ET
from musicxml.xmlelement.xmlelement import *


def _et_xml_to_music_xml(node):
    output = eval(convert_to_xml_class_name(node.tag))()
    for k, v in node.attrib.items():
        try:
            setattr(output, k, v)
        except (TypeError, ValueError):
            try:
                setattr(output, k, int(v))
            except ValueError:
                setattr(output, k, float(v))

    if node.text:
        text = node.text.strip()
        if text:
            try:
                output.value = text
            except TypeError:
                try:
                    output.value = int(text)
                except ValueError:
                    output.value = float(text)
    return output


def _parse_node(xml_node):
    output = _et_xml_to_music_xml(xml_node)
    for child in xml_node:
        output.add_child(_parse_node(child))
    return output


def parse_musicxml(file_path):
    with open(file_path) as file:
        xml = ET.parse(file)
    return _parse_node(xml.getroot())
